/* date.c - parse formatted times
   Copyright (C) 1996-2000 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mostincludes.h"
#include "mad.h"

#define swap(x,y) {int __t; __t = (x); (x) = (y); (y) = __t;}

/* parse date strings like: Sun Oct 31 04:47:07 1999 */

static int subtract_string (char *str, char *subtraction)
{
    char *t;
    int found = 0;
    while ((t = strstr (str, subtraction))) {
	found = 1;
	strcpy (t, t + strlen (subtraction));
	while (*t && isalpha (*t))
	    strcpy (t, t + 1);
    }
    return found;
}

static void make_lower (char *s)
{
    while (*s)
	*s++ = tolower (*s);
}

static void chomp (char *r)
{
    char *p;
    while (*r && isspace (*r))
	strcpy (r, r + 1);
    for (p = r + strlen (r) - 1; (unsigned long) p > (unsigned long) r; p--)
	if (isspace (*p))
	    *p = '\0';
	else
	    break;
}


#define DIGITS "0123456789"

time_t get_time_from_date_string (char *s, char **result_str)
{
    char weekday[8][8];
    char month[13][8];
    char fmt[256], am[8], pm[8];
    char *r, *q, *e;
    int n, offset = 0, found_month = 0, wday, found_weekday = 0, i;
    struct tm t;
    time_t result;

/* set to lower case */
    r = strdup (s);
    make_lower (r);

/* get weekday names for current locale */
    memset (&t, 0, sizeof (t));
    for (i = 0; i < 7; i++) {
	t.tm_wday = i;
	strftime (weekday[i], 7, "%a", &t);
	make_lower (weekday[i]);
    }
    weekday[i][0] = '\0';

/* get month names for current locale */
    for (i = 0; i < 12; i++) {
	t.tm_mon = i;
	strftime (month[i], 7, "%b", &t);
	make_lower (month[i]);
    }
    month[i][0] = '\0';

/* get am/pm */
    t.tm_hour = 6;
    strftime (am, 7, "%p", &t);
    make_lower (am);
    t.tm_hour = 18;
    strftime (pm, 7, "%p", &t);
    make_lower (pm);

/* clear time struct */
    memset (&t, 0, sizeof (t));

/* get weekday */
    for (wday = 0; weekday[wday][0]; wday++)
	if (subtract_string (r, weekday[wday])) {
	    found_weekday = 1;
	    break;
	}

/* get month */
    for (t.tm_mon = 0; month[t.tm_mon][0]; t.tm_mon++)
	if (subtract_string (r, month[t.tm_mon])) {
	    found_month = 1;
	    break;
	}

/* should not have both these conditions */
    if (found_weekday && !found_month)
	return 0;

/* ... but a date like ` 1999 ' is ok */
    if (!found_month)
	t.tm_mon = 0;

/* get time */
    for (n = 0, q = r; *q; q++) {
	if (sscanf (q, "%u:%u:%u%n", &t.tm_hour, &t.tm_min, &t.tm_sec, &n) >= 3)
	    break;
	if (sscanf (q, "%uh%u:%u%n", &t.tm_hour, &t.tm_min, &t.tm_sec, &n) >= 3)
	    break;
	if (sscanf (q, "%u:%u%n", &t.tm_hour, &t.tm_min, &n) >= 2)
	    break;
	if (sscanf (q, "%uh%u%n", &t.tm_hour, &t.tm_min, &n) >= 2)
	    break;
    }
    if (n) {
	char *p;
	strcpy (q, q + n);

/* get am/pm */
	if ((p = strstr (q, am))) {
	    strcpy (p, p + strlen (am));
	} else if ((p = strstr (q, pm))) {
	    strcpy (p, p + strlen (pm));
	    t.tm_hour += 12;
	}

/* get meridian */
	for (n = 0, q = r; *q; q++) {
	    if (sscanf (q, "+%u%n", &offset, &n) >= 1)
		break;
	    if (sscanf (q, "-%u%n", &offset, &n) >= 1) {
		offset = -offset;
		break;
	    }
	}
	strcpy (q, q + n);
    } else {
/* assume 00:00:00 */
	t.tm_min = t.tm_sec = 0;
	t.tm_hour = 0;
    }

/* get month */
    q = r + strcspn (r, DIGITS);
    if (*q) {
	t.tm_mday = strtol (q, &e, 10);
	strcpy (q, e);
    } else {
	t.tm_mday = 1;
    }

/* get year */
    q = r + strcspn (r, DIGITS);
    if (*q) {
	t.tm_year = strtol (q, &e, 10);
	strcpy (q, e);
/* correct year */
	if (t.tm_year < t.tm_mday)
	    swap (t.tm_year, t.tm_mday);
	if (t.tm_year > 100)
	    t.tm_year -= 1900;
    } else {
/* default to current year if ommitted */
	time_t lt;
	time (&lt);
	t.tm_year = localtime (&lt)->tm_year;
    }

/* get time */
    result = mktime (&t);
    if (found_weekday)
	if (t.tm_wday != wday)
	    return 0;
    if (offset) {
	if (abs (offset) > 15)
	    result -= timezone + (offset / 100) * 60 * 60 + (offset % 100) * 60;
	else
	    result -= timezone + offset * 60 * 60;
    }
    chomp (r);
    if (result_str)
	*result_str = strdup (r);

#if 0
    strftime (fmt, 256, "%a, %d %b %Y %H:%M:%S %z", localtime (&result));
    printf ("%s --> %s : %s\n", s, fmt, r);
#endif

    free (r);
    return result;
}



