#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/file.h>
#include <errno.h>

#ifndef MAXINT
#define MAXINT ((int)(~0U>>1))
#endif

#define BUFSIZ 1024

static int check_link_count (const char *file)
{
    struct stat sb;
    if (stat (file, &sb))
	return 0;
    if (sb.st_nlink != 2) {
	return 0;
    }
    unlink (file);
    return 1;
}

/* return non-zero on success */
static int do_lock_file (const char *file, const char *lock)
{
    int fd;
    int pid;
    int len;
    char buf[32];
    /*
     * Create a lock file which can be switched into place
     */
    if ((fd = open (file, O_CREAT | O_EXCL | O_WRONLY, 0600)) == -1) {
	return 0;
    }
    pid = getpid ();
    sprintf (buf, "%d", pid);
    len = strlen (buf) + 1;
    if (write (fd, buf, len) != len) {
	(void) close (fd);
	(void) unlink (file);
	return 0;
    }
    close (fd);
    /*
     * Simple case first -
     *      Link fails (in a sane environment ...) if the target
     *      exists already.  So we try to switch in a new lock
     *      file.  If that succeeds, we assume we have the only
     *      valid lock.  Needs work for NFS where this assumption
     *      may not hold.  The simple hack is to check the link
     *      count on the source file, which should be 2 iff the
     *      link =really= worked.
     */
    if (link (file, lock) == 0)
	return check_link_count (file);
    /*
     * Invalid lock test -
     *      Open the lock file and see if the lock is valid.
     *      The PID of the lock file is checked, and if the PID
     *      is not valid, the lock file is removed.  If the unlink
     *      of the lock file fails, it should mean that someone
     *      else is executing this code.  They will get success,
     *      and we will fail.
     */
    if ((fd = open (lock, O_RDWR)) == -1 || (len = read (fd, buf, sizeof buf)) <= 0) {
	errno = EINVAL;
	return 0;
    }
    buf[len] = '\0';
    if ((pid = strtol (buf, (char **) 0, 10)) == 0) {
	errno = EINVAL;
	return 0;
    }
    if (kill (pid, 0) == 0) {
	errno = EEXIST;
	unlink (file);
	return 0;
    }
    fprintf (stderr, "mailgrep: removing stale lock file\n");
    if (unlink (lock)) {
	(void) close (fd);
	(void) unlink (file);
	return 0;
    }
    /*
     * Re-try lock -
     *      The invalid lock has now been removed and I should
     *      be able to acquire a lock for myself just fine.  If
     *      this fails there will be no retry.  The link count
     *      test here makes certain someone executing the previous
     *      block of code didn't just remove the lock we just
     *      linked to.
     */
    if (link (file, lock) == 0)
	return check_link_count (file);
    (void) unlink (file);
    return 0;
}

/* perform hardlink *.lock style locking */
/* returns lock fd on success (which must be passed to unlock_file) */
int lock_file (char *filename)
{
    char file[BUFSIZ], lock[BUFSIZ], lock2[BUFSIZ];
    if (strlen (filename) > 1000)
	return -1;
    sprintf (file, "%s.%d", filename, getpid ());
    sprintf (lock, "%s.lock", filename);
    if (do_lock_file (file, lock))
	return 0;
    return -1;
}

void unlock_file (char *filename)
{
    char lock[BUFSIZ];
    if (strlen (filename) > 1000)
	return;
    sprintf (lock, "%s.lock", filename);
    unlink (lock);
}

