/* mailgrep.c - main file
   Copyright (C) 1996, 1997 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mostincludes.h"
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "parse.h"
#include "regex.h"
#include "cmdlineopt.h"
#include "mad.h"

static char *progname = "";
static Value heap[256];
static void *compiled_script = 0;
static char **extra_args = 0;

static int show_version = 0;
static int verbose = 0;
static int show_help = 0;
static int option_test = 0;
static long max_mem = 1024 * 1024;

#define MAX_MSG_GREP VALUE_LEN

time_t get_time_from_date_string (char *str, char **str_result);

static char *current_message = 0;
static long current_alloced = 0;

static char *output_file = 0;

void unlock_file (char *filename);
int lock_file (char *filename);

/* altimate rename */
int my_rename (char *f, char *g)
{
    unsigned char buf[1024];
    int fd, gd;
    struct stat st1, st2;
    memset (&st1, 0, sizeof (st1));
    memset (&st2, 0, sizeof (st2));
    if (stat (f, &st1))
	return 1;
    if ((gd = open (g, O_WRONLY | O_TRUNC | O_CREAT, st1.st_mode)) < 0)
	return 1;
    close (gd);
    if (stat (g, &st2))
	return 1;
    if (st1.st_dev == st2.st_dev) {
	if (verbose)
	    fprintf (stderr, "%s: %s and %s on same device, renaming\n", progname, f, g);
	close (gd);
	return rename (f, g);
    }
    if ((gd = open (g, O_WRONLY | O_TRUNC | O_CREAT, st1.st_mode)) < 0)
	return 1;
    if ((fd = open (f, O_RDONLY)) < 0) {
	close (gd);
	return 1;
    }
    for (;;) {
	int c, r = 0;
	unsigned char *p;
	c = read (fd, buf, 1024);
	if (!c)
	    break;
	if (c < 0) {
	    close (fd);
	    close (gd);
	    return -1;
	}
	for (p = buf; c > 0; p += r) {
	    int r;
	    r = write (gd, p, c);
	    if (r <= 0) {
		close (fd);
		close (gd);
		return 1;
	    }
	    c -= r;
	}
    }
    return close (fd) | close (gd) | unlink (f);
}

#define mail_free(x) 							\
    { 									\
	if (x) {							\
	    free (x);							\
	    (x) = 0;							\
	}								\
    }

/* efficient fgets */
#ifdef HAVE_MAD
static int mad_read_one_line (char **line, int f, char *file, int line_)
#define read_one_line(l,f) mad_read_one_line(l,f,__FILE__,__LINE__)
#else
static int read_one_line (char **line, int f)
#endif
{
    static unsigned char buf[8192];
    static unsigned char *bufp = buf;
    static int buflen = 0;
    int alloced, resultlen = 0;
    unsigned char *result;
#ifdef HAVE_MAD
    result = (unsigned char *) mad_alloc (alloced = 256, file, line_);
#else
    result = (unsigned char *) malloc (alloced = 256);
#endif
    for (;;) {
	int t;
	unsigned char *newline;
	t = buflen;
	if ((newline = memchr (bufp, '\n', t)))
	    t = (unsigned long) newline - (unsigned long) bufp + 1;
	while (resultlen + t >= alloced - 2)
#ifdef HAVE_MAD
	    result = (unsigned char *) mad_realloc (result, alloced <<= 1, file, line_);
#else
	    result = (unsigned char *) realloc (result, alloced <<= 1);
#endif
	memcpy (result + resultlen, bufp, t);
	resultlen += t;
	result[resultlen] = '\0';
	buflen -= t;
	bufp += t;
	if (newline)
	    break;
	if (!buflen)
	    bufp = buf;
	while ((t = read (f, bufp, 8192 - ((unsigned long) bufp - (unsigned long) buf))) == -1 && errno == EINTR);
	if (t <= 0) {
	    free (result);
	    return 0;
	}
	buflen += t;
    }
    if (resultlen)
	result[resultlen - 1] = '\0';
    *line = result;
    return resultlen;
}

static struct message_list *messages = 0;

struct header_list {
    char *line;
    struct header_list *prev;
    struct header_list *next;
};

struct message_list {
    long start;
    long end;
    struct message_list *prev;
    struct message_list *next;
};

#define REWIND(l)						\
		    {						\
			if (l)					\
			    while ((l)->prev)			\
				(l) = (l)->prev;		\
		    }

void free_message_list (struct message_list **messages)
{
    struct message_list *p, *t;
    REWIND ((*messages));
    for (p = *messages; p; p = t) {
	t = p->next;
	free (p);
    }
    *messages = 0;
}

static void omit_message (struct message_list **l, int offset1, int offset2)
{
    if (*l) {
	if ((*l)->end == offset1) {	/* consecutive oversized messages */
	    (*l)->end = offset2;
	    return;
	}
	(*l)->next = malloc (sizeof (struct message_list));
	(*l)->next->prev = *l;
	(*l) = (*l)->next;
    } else {
	(*l) = malloc (sizeof (struct message_list));
	(*l)->prev = 0;
    }
    (*l)->start = offset1;
    (*l)->end = offset2;
    (*l)->next = 0;
}

static void append_header_list (struct header_list **h, char *p)
{
    if (*h) {
	if (*p == '\t') {	/* line continuation */
	    char *s;
	    s = malloc (strlen ((*h)->line) + strlen (p) + 2);
	    strcpy (s, (*h)->line);
	    strcat (s, "\n");
	    strcat (s, p);
	    free ((*h)->line);
	    (*h)->line = s;
	    return;
	}
	(*h)->next = malloc (sizeof (struct header_list));
	(*h)->next->prev = *h;
	(*h) = (*h)->next;
    } else {
	(*h) = malloc (sizeof (struct header_list));
	(*h)->prev = 0;
    }
    (*h)->line = strdup (p);
    (*h)->next = 0;
}

static void free_header_list (struct header_list **h)
{
    struct header_list *s, *t;
    REWIND (*h);
    for (s = *h; s; s = t) {
	t = s->next;
	mail_free (s->line);
	mail_free (s);
    }
    *h = 0;
}

/* result must be free'd */
static char *find_header_line (struct header_list *h, char *p)
{
    struct header_list *s;
    int l;
    l = strlen (p);
    for (s = h; s; s = s->next) {
	if (!strncmp (s->line, p, l)) {
	    char *r;
	    r = malloc (strlen (s->line) - l + 1);
	    strcpy (r, s->line + l);
	    return r;
	}
    }
    return 0;
}

struct message_info {
    struct message_list **l;
    struct header_list *h;
    time_t message_time;
    int message_length;
    int header_length;
    char **message;
    long offset;
};

static int run_script (struct message_list **l, struct header_list *h, char **message, long offset, int message_length,
		       int header_length, time_t message_time)
{
    struct message_info i;
    i.l = l;
    i.h = h;
    i.message_length = message_length;
    i.header_length = header_length;
    i.offset = offset;
    i.message = message;
    i.message_time = message_time;
    return parser_evaluate (compiled_script, &i);
}

static time_t get_time_from_From_field (char *s)
{
    s += strspn (s, "From ");
    while (!isspace (*s))
	s++;
    return get_time_from_date_string (s, 0);
}

static int scan_file (FILE * f)
{
    long offset = 0, last_seek = 0, r, lastlinelength = 1, seek;
    int header_length = 0;
    int msg_no = 0;
    long message_read = 0;
    struct header_list *h = 0;
    time_t message_time;

    current_message = malloc (current_alloced = 65536);
    current_message[0] = '\0';
    do {
	char *s;
	seek = offset;
	offset += (r = read_one_line (&s, fileno (f)));

/* EOF or start of header */
	if (!r || (!strncmp (s, "From ", 5) && lastlinelength == 1)) {
	    int return_value;

/* run script on last message */
	    if (h) {
		REWIND (h);
		if (verbose)
		    printf ("%d\n", ++msg_no);
		return_value =
		    run_script (&messages, h, &current_message, last_seek, seek - last_seek - header_length,
				header_length, message_time);
	    }

/* reset message body */
	    message_read = 0;
	    if (current_alloced > 65536)
		current_message = realloc (current_message, current_alloced = 65536);
	    current_message[0] = '\0';

/* free header */
	    free_header_list (&h);
	    header_length = 0;

/* remember offset into mail folder */
	    last_seek = seek;

/* break if EOF */
	    if (!r) {
		mail_free (s);
		break;
	    }

/* start new header */
	    append_header_list (&h, s);
	    message_time = get_time_from_From_field (s);
	    mail_free (s);

#ifdef HAVE_MAD
	    mad_count (__FILE__, __LINE__);
#endif

/* while none-blank line, read entire header */
	    while (r > 1) {
		char *p;
		offset += (r = read_one_line (&p, fileno (f)));
		if (r > 1)
		    append_header_list (&h, p);
		mail_free (p);
	    }
	    header_length = offset - seek;
	} else {
/* message body line */
/* first resize pool if needed */
	    while (message_read + r >= current_alloced - 2 && current_alloced < max_mem) {
		current_alloced <<= 1;
		current_message = realloc (current_message, current_alloced);
	    }
	    if (message_read + r < current_alloced - 2) {
		memcpy (current_message + message_read, s, r);
		message_read += r - 1;
		current_message[message_read++] = '\n';
		current_message[message_read] = '\0';
	    }
	    mail_free (s);
	}
	lastlinelength = r;
    } while (r);
    mail_free (current_message);
    return 0;
}

static int rewrite_file (FILE * f, FILE * g)
{
    int r;
    long seek, offset = 0;
    REWIND (messages);
    do {
	char *s;
	seek = offset;
	offset += (r = read_one_line (&s, fileno (f)));
	if (messages)
	    if (seek > messages->end)
		messages = messages->next;
	if (r) {
	    if (messages) {
		if (seek < messages->start || seek >= messages->end) {
		    fputs (s, g);
		    fputs ("\n", g);
		}
	    } else {
		fputs (s, g);
		fputs ("\n", g);
	    }
	}
	mail_free (s);
    } while (r);
    return 0;
}

/* {{{ functions */

static int op_regexp (void *dummy, ValueStack ** v)
{
/* FIXME: not thread safe */
    static regex_t regexp;
    Value value1, value2;
    char *p = 0;
    static char *s = 0;
    value2 = parser_value_pop (v);
    value1 = parser_value_pop (v);
    if (!((value2.type & value1.type) & VALUE_STRING)) {
	fprintf (stderr, "%s: bad arg type to regexp()\n", progname);
	return 1;
    }
    if (!s) {
	goto new_comp;
	memset (&regexp, 0, sizeof (regexp));	/* probably don't need this */
    }
    if (strcmp (s, value1.v.s)) {
      new_comp:
	p = value1.v.s;
	if (s) {
	    free (s);
	    regfree (&regexp);
	}
	s = (char *) strdup (value1.v.s);
	if (regcomp (&regexp, p, REG_EXTENDED | REG_NOSUB)) {
	    fprintf (stderr, "%s: regexp(): error compiling regular expression", progname);
	    return 1;
	}
    }
    parser_push_int (v, regexec (&regexp, value2.v.s, 0, 0, 0));
    parser_value_free (value1);
    parser_value_free (value2);
    return 0;
}

static int op_field (struct message_info *i, ValueStack ** v)
{
    char *s;
    Value value;
    value = parser_value_pop (v);
    if (!(value.type & VALUE_STRING)) {
	fprintf (stderr, "%s: bad arg type to field()\n", progname);
	return 1;
    }
    s = find_header_line (i->h, value.v.s);
    if (s)
	parser_push_string (v, s, strlen (s));
    else
	parser_push_int (v, 0);
    parser_value_free (value);
    return 0;
}

static int op_message_length (struct message_info *i, ValueStack ** v)
{
    parser_push_int (v, i->message_length);
    return 0;
}

static int op_header_length (struct message_info *i, ValueStack ** v)
{
    parser_push_int (v, i->header_length);
    return 0;
}

static int op_offset (struct message_info *i, ValueStack ** v)
{
    parser_push_int (v, i->offset);
    return 0;
}

static int op_message (struct message_info *i, ValueStack ** v)
{
    parser_push_string (v, *(i->message), strlen (*(i->message)));
    return 0;
}

static int op_header (struct message_info *i, ValueStack ** v)
{
    struct header_list *s;
    char *p = 0;
    int l = 0;
    for (s = i->h; s; s = s->next) {
	int len;
	char *q;
	len = strlen (s->line) + 1;
	q = malloc (l + len);
	if (p) {
	    memcpy (q, p, l);
	    free (p);
	}
	p = q;
	if (l)
	    p[l - 1] = '\n';
	strcpy (p + l, s->line);
	l += len;
    }
    p[l - 1] = '\0';
    parser_push_string_free (v, p, l);
    return 0;
}

/* parse date strings like: Sun Oct 31 04:47:07 1999 */
static int op_parsedate (struct message_info *i, ValueStack ** v)
{
    Value value;
    long t;
    value = parser_value_pop (v);
    if (!(value.type & VALUE_STRING)) {
	runtime_error ("non-string arg to parsedate()", 0);
	return 1;
    }
    t = (long) get_time_from_date_string (value.v.s, 0);
    if (!t || t == -1) {
	runtime_error ("cannot parse date string", 0);
	return 1;
    }
    parser_push_int (v, t);
    return 0;
}

static int op_date (struct message_info *i, ValueStack ** v)
{
    parser_push_int (v, (long) i->message_time);
    return 0;
}

static int op_delete (struct message_info *i, ValueStack ** v)
{
    omit_message (i->l, i->offset, i->offset + i->header_length + i->message_length);
    parser_push_int (v, i->offset);
    return 0;
}

static Operator operators[] = {
    {0, "OFFSET", 0, OP_PREDEF, 1, (Func) op_offset, 0, 0},
    {0, "HEADER_LENGTH", 0, OP_PREDEF, 1, (Func) op_header_length, 0, 0},
    {0, "MESSAGE_LENGTH", 0, OP_PREDEF, 1, (Func) op_message_length, 0, 0},
    {0, "field", 0, OP_USER_FUNCTION, 1, (Func) op_field, 0, 1},
    {0, "regexp", 0, OP_USER_FUNCTION, 1, (Func) op_regexp, 0, 2},
    {0, "delete", 0, OP_USER_FUNCTION, 1, (Func) op_delete, 0, 0},
    {0, "message", 0, OP_USER_FUNCTION, 1, (Func) op_message, 0, 0},
    {0, "header", 0, OP_USER_FUNCTION, 1, (Func) op_header, 0, 0},
    {0, "date", 0, OP_USER_FUNCTION, 1, (Func) op_date, 0, 0},
    {0, "parsedate", 0, OP_USER_FUNCTION, 1, (Func) op_parsedate, 0, 1},
};

static int n_operator = sizeof (operators) / sizeof (Operator);

/* }}} functions */

static int start ();

static int ummm_start (void)
{
    int r = 0;
    char *temp_file;

/* lock file */
    while (lock_file (extra_args[0]))
	sleep (1);

    if (lock_file (temp_file = tempnam ("/tmp", "mlgrp"))) {
	unlock_file (extra_args[0]);
	fprintf (stderr, "%s: cannot lock temp file %s\n", progname, temp_file);
	return 1;
    }

    unlink (temp_file);

/* process the mailbox */
    r = start (temp_file);
    free_message_list (&messages);

/* unlock file */
    unlock_file (extra_args[0]);
    unlock_file (temp_file);
    free (temp_file);
    return r;
}

static int start (char *temp_file)
{
    FILE *f = 0, *g = 0;
    int r, fd;
    compiled_script = parser_compile (extra_args[1], heap);
    if (!compiled_script)
	return 1;
    fd = open (extra_args[0], O_RDONLY);
    if (fd == -1) {
	fprintf (stderr, "%s: %s: %s\n", progname, extra_args[0], strerror (errno));
	return 1;
    }
    f = fdopen (fd, "r");
    r = scan_file (f);
    parser_free (compiled_script, heap);
    fclose (f);
/* dummy `while' for `break's */
    if (!r && (messages || output_file)) {
	struct stat st;
	if (stat (extra_args[0], &st)) {
	    fprintf (stderr, "%s: %s: %s\n", progname, extra_args[0], strerror (errno));
	    return 1;
	}
	f = fopen (extra_args[0], "r");
	if (!f) {
	    fprintf (stderr, "%s: %s: %s\n", progname, extra_args[0], strerror (errno));
	    return 1;
	}
	fd = open (temp_file, O_WRONLY | O_TRUNC | O_CREAT, st.st_mode);
	if (fd < 0) {
	    fprintf (stderr, "%s: %s: %s\n", progname, temp_file, strerror (errno));
	    return 1;
	}
	g = fdopen (fd, "w");
	if (!g) {
	    fprintf (stderr, "%s: %s: %s\n", progname, temp_file, strerror (errno));
	    return 1;
	}
	rewrite_file (f, g);
	fclose (f);
	fclose (g);
	if (!option_test) {
	    if (output_file) {
		if (my_rename (temp_file, output_file)) {
		    fprintf (stderr, "%s: error moving temp file to %s: %s\n", progname, output_file, strerror (errno));
		    return 1;
		}
	    } else {
		unlink (extra_args[0]);
		if (my_rename (temp_file, extra_args[0])) {
		    fprintf (stderr,
			     "%s: error moving temp file to %s: %s\n", progname, extra_args[0], strerror (errno));
		    return 1;
		}
	    }
	}
    }
    return 0;
}

static void usage (void)
{
    fprintf (stderr, "\
Usage:\n\t%s -[vVh] <mailbox> '<script>'\n\
Scans through file <mailbox>, executing <script> on each mail message.\n\
\n\
Options:\n\
 -v, --verbose         verbose output\n\
 -h, --help            print this help message\n\
 -V, --version         print version number and exit\n\
 -t, --dry-run         test only, do not modify the mailbox\n\
 -o, --output          use different output file (do not modify mailbox)\n\
 -m, --max-mem         max memory to allocate for message body\n\
", progname);
    exit (1);
}

static int pars_opts (int argc, char **argv)
{
    int i;
    struct prog_options arg_opts[] = {
	{' ', "", "", ARG_STRINGS, 0, 0, 0},
	{'V', "", "--version", ARG_SET, 0, 0, &show_version},
	{'v', "", "--verbose", ARG_ADD, 0, 0, &verbose},
	{'h', "", "--help", ARG_SET, 0, 0, &show_help},
	{'t', "", "--dry-run", ARG_SET, 0, 0, &option_test},
	{'o', "", "--output", ARG_STRING, &output_file, 0, 0},
	{'m', "", "--max-mem", ARG_INT, 0, 0, &max_mem},
	{0, 0, 0, 0, 0, 0, 0}
    };
    extra_args = malloc (argc * sizeof (char *));
    arg_opts[0].strs = extra_args;
    memset (extra_args, 0, argc * sizeof (char *));
    i = get_cmdline_options (argc, argv, arg_opts);
    if (i) {
	fprintf (stderr, "%s: error on cmdline near: %s", progname, argv[i]);
	return 1;
    }
    if (show_help) {
	usage ();
	exit (0);
    }
    if (show_version) {
	printf ("Mailgrep version " VERSION "\n");
	exit (0);
    };
    if (extra_args[2]) {
	fprintf (stderr, "%s: error on cmdline near: %s", progname, extra_args[2]);
	return 1;
    }
    if (!extra_args[1]) {
	fprintf (stderr, "%s: messing args on cmdline", progname);
	return 1;
    }
    return 0;
}

#if (RETSIGTYPE==void)
#define handler_return return
#else
#define handler_return return 1
#endif

#if 0
/* must remove the lock file if we recieve a singal */
static RETSIGTYPE quit_handler (int x)
{
    if (temp_file)
	unlink (temp_file);
    unlock_file (extra_args[0]);
    exit (1);
    handler_return;
}
#endif

int main (int argc, char **argv)
{
    int r, i;
    char *p;
    p = strrchr (argv[0], '/');
    progname = strdup (p ? p + 1 : argv[0]);
    if (pars_opts (argc, argv)) {
	usage ();
	exit (1);
    }

    for (i = 0; i < n_operator; i++)
	parser_add_operator (operators + i);
    parser_init ();
    r = ummm_start ();
    parser_shut ();
    free (progname);
    free (extra_args);

#ifdef HAVE_MAD
    printf ("calling mad_finalize ()\n");
    mad_finalize (__FILE__, __LINE__);
#endif

    exit (r);
    return r;
}


